Package: polite
Version: 0.1.4
Title: Be Nice on the Web
Description: Be responsible when scraping data from websites by following polite principles: introduce yourself, ask for permission, take slowly and never ask twice.
Authors@R: person(given="Dmytro", family="Perepolkin", email = "dperepolkin@gmail.com", role = c("aut", "cre"), comment = c(ORCID="0000-0001-8558-6183"))
License: MIT + file LICENSE
Encoding: UTF-8
ByteCompile: true
URL: https://github.com/dmi3kno/polite,
        https://dmi3kno.github.io/polite/
BugReports: https://github.com/dmi3kno/polite/issues
Imports: httr, magrittr, memoise, ratelimitr, robotstxt, rvest, stats,
        usethis
Suggests: dplyr, testthat, covr, webmockr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-10 19:50:22 UTC; DmytroPerepolkin
Author: Dmytro Perepolkin [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-8558-6183>)
Maintainer: Dmytro Perepolkin <dperepolkin@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-11 05:10:02 UTC
Built: R 4.5.2; ; 2026-05-11 06:41:36 UTC; unix

Package: polite
Version: 0.1.4
Title: Be Nice on the Web
Description: Be responsible when scraping data from websites by following polite principles: introduce yourself, ask for permission, take slowly and never ask twice.
Authors@R: person(given="Dmytro", family="Perepolkin", email = "dperepolkin@gmail.com", role = c("aut", "cre"), comment = c(ORCID="0000-0001-8558-6183"))
License: MIT + file LICENSE
Encoding: UTF-8
ByteCompile: true
URL: https://github.com/dmi3kno/polite,
        https://dmi3kno.github.io/polite/
BugReports: https://github.com/dmi3kno/polite/issues
Imports: httr, magrittr, memoise, ratelimitr, robotstxt, rvest, stats,
        usethis
Suggests: dplyr, testthat, covr, webmockr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-10 19:50:22 UTC; DmytroPerepolkin
Author: Dmytro Perepolkin [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-8558-6183>)
Maintainer: Dmytro Perepolkin <dperepolkin@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-11 05:10:02 UTC
Built: R 4.5.2; ; 2026-08-22 14:48:57 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,polite)
export("%>%")
export(bow)
export(guess_basename)
export(html_attrs_dfr)
export(is.polite)
export(nod)
export(politely)
export(rip)
export(scrape)
export(set_rip_delay)
export(set_scrape_delay)
export(use_manners)
importFrom(httr,GET)
importFrom(httr,HEAD)
importFrom(httr,add_headers)
importFrom(httr,config)
importFrom(httr,content)
importFrom(httr,handle)
importFrom(httr,headers)
importFrom(httr,http_error)
importFrom(httr,modify_url)
importFrom(httr,parse_url)
importFrom(magrittr,"%>%")
importFrom(memoise,cache_memory)
importFrom(memoise,forget)
importFrom(memoise,has_cache)
importFrom(memoise,memoise)
importFrom(ratelimitr,UPDATE_RATE)
importFrom(ratelimitr,get_rates)
importFrom(ratelimitr,limit_rate)
importFrom(ratelimitr,rate)
importFrom(robotstxt,robotstxt)
importFrom(rvest,html_attr)
importFrom(rvest,html_attrs)
importFrom(rvest,html_text)
importFrom(stats,na.omit)
importFrom(stats,setNames)
importFrom(tools,file_ext)
importFrom(usethis,use_template)
importFrom(utils,download.file)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# polite <img src="man/figures/logo.png" align="right" />

<!-- badges: start -->

[![AppVeyor build
status](https://ci.appveyor.com/api/projects/status/github/dmi3kno/polite?branch=master&svg=true)](https://ci.appveyor.com/project/dmi3kno/polite)
[![Codecov test
coverage](https://codecov.io/gh/dmi3kno/polite/branch/master/graph/badge.svg)](https://app.codecov.io/gh/dmi3kno/polite?branch=master)
[![CRAN
status](https://www.r-pkg.org/badges/version/polite)](https://CRAN.R-project.org/package=polite)
[![Lifecycle:
maturing](https://img.shields.io/badge/lifecycle-maturing-blue.svg)](https://lifecycle.r-lib.org/articles/stages.html#maturing)
[![R-CMD-check](https://github.com/dmi3kno/polite/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/dmi3kno/polite/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

The goal of `polite` is to promote responsible web etiquette.

> **“bow and scrape” (verb):**
>
> 1)  To make a deep bow with the right leg drawn back (thus scraping
>     the floor), left hand pressed across the abdomen, right arm held
>     aside.
>
> 2)  *(idiomatic, by extension)* To behave in a servile, obsequious, or
>     excessively polite manner. \[1\] Source: *Wiktionary, The free
>     dictionary*

The package’s two main functions `bow` and `scrape` define and realize a
web harvesting session. `bow` is used to introduce the client to the
host and ask for permission to scrape (by inquiring against the host’s
`robots.txt` file), while `scrape` is the main function for retrieving
data from the remote server. Once the connection is established, there’s
no need to `bow` again. Rather, in order to adjust a scraping URL the
user can simply `nod` to the new path, which updates the session’s URL,
making sure that the new location can be negotiated against
`robots.txt`.

The three pillars of a `polite session` are **seeking permission, taking
slowly and never asking twice**.

The package builds on awesome toolkits for defining and managing http
sessions (`httr` and `rvest`), declaring the user agent string and
investigating site policies (`robotstxt`), and utilizing rate-limiting
and response caching (`ratelimitr` and `memoise`).

## Installation

You can install `polite` from [CRAN](https://cran.r-project.org/) with:

``` r
install.packages("polite")
```

Development version of the package can be installed from
[Github](https://github.com/dmi3kno/polite) with:

``` r
install.packages("remotes")
remotes::install_github("dmi3kno/polite")
```

## Basic Example

This is a basic example which shows how to retrieve the list of
semi-soft cheeses from www.cheese.com. Here, we authenticate a session
and then scrape the page with specified parameters. Behind the scenes
`polite` retrieves `robots.txt`, checks the URL and user agent string
against it, caches the call to `robots.txt` and to the web page and
enforces rate limiting.

``` r
library(polite)
library(rvest)

session <- bow("https://www.cheese.com/by_type", force = TRUE)
result <- scrape(session, query=list(t="semi-soft", per_page=100)) %>%
  html_node("#main-body") %>%
  html_nodes("h3") %>%
  html_text()
head(result)
#> [1] "American Cheese"     "Mozzarella"          "Taleggio"           
#> [4] "Fontina Val d'Aosta" "Blue Cheese"         "Jarlsberg"
```

## Extended Example

You can build your own functions that incorporate `bow`, `scrape` (and,
if required, `nod`). Here we will extend our inquiry into cheeses and
will download all cheese names and URLs to their information pages.
Let’s retrieve the number of pages per letter in the alphabetical list,
keeping the number of results per page to 100 to minimize number of web
requests.

``` r
library(polite)
library(rvest)
library(purrr)
library(dplyr)

session <- bow("https://www.cheese.com/alphabetical")

# this is only to illustrate the example.
letters <- letters[1:3] # delete this line to scrape all letters

responses <- map(letters, ~scrape(session, query = list(per_page=100,i=.x)) )
results <- map(responses, ~html_nodes(.x, "#id_page li") %>%
                           html_text(trim = TRUE) %>%
                           as.numeric() %>%
                           tail(1) ) %>%
           map(~pluck(.x, 1, .default=1))
pages_df <- tibble(letter = rep.int(letters, times=unlist(results)),
                   pages = unlist(map(results, ~seq.int(from=1, to=.x))))
pages_df
#> # A tibble: 3 × 2
#>   letter pages
#>   <chr>  <int>
#> 1 a          1
#> 2 b          1
#> 3 c          1
```

Now that we know how many pages to retrieve from each letter page, let’s
rotate over letter pages and retrieve cheese names and underlying links
to cheese details. We will need to write a helper function. Our session
is still valid and we don’t need to `nod` again, because we will not be
modifying a page URL, only its parameters (note that the field `url` is
missing from `scrape` function).

``` r
get_cheese_page <- function(letter, pages){
 lnks <- scrape(session, query=list(per_page=100,i=letter,page=pages)) %>%
    html_nodes("h3 a")
tibble(name=lnks %>% html_text(),
       link=lnks %>% html_attr("href"))
}

df <- pages_df %>% pmap_df(get_cheese_page)
df
#> # A tibble: 60 × 2
#>    name                    link                     
#>    <chr>                   <chr>                    
#>  1 Aarewasser              /aarewasser/             
#>  2 Abbaye de Belloc        /abbaye-de-belloc/       
#>  3 Abbaye de Belval        /abbaye-de-belval/       
#>  4 Abbaye de Citeaux       /abbaye-de-citeaux/      
#>  5 Abbaye de Tamié         /tamie/                  
#>  6 Abbaye de Timadeuc      /abbaye-de-timadeuc/     
#>  7 Abbaye du Mont des Cats /abbaye-du-mont-des-cats/
#>  8 Abbot’s Gold            /abbots-gold/            
#>  9 Abertam                 /abertam/                
#> 10 Abondance               /abondance/              
#> # ℹ 50 more rows
```

## Another example

Bob Rudis is one the vocal proponents of an online etiquette in the R
community. If you have never seen his robots.txt file, you should
definitely [check it out](https://rud.is/robots.txt)! Lets look at his
[blog](https://rud.is/b/). We don’t know how many pages will the gallery
return, so we keep going until there’s no more “Older posts” button.
Note that I first `bow` to the host and then simply `nod` to the current
scraping page inside the `while` loop.

``` r
    library(polite)
    library(rvest)

    hrbrmstr_posts <- data.frame()
    url <- "https://rud.is/b/"
    session <- bow(url)

    while(!is.na(url)){
      # make it verbose
      message("Scraping ", url)
      # nod and scrape
      current_page <- nod(session, url) %>%
        scrape(verbose=TRUE)
      # extract post titles
      hrbrmstr_posts <- current_page %>%
        html_nodes(".entry-title a") %>%
        polite::html_attrs_dfr() %>%
        rbind(hrbrmstr_posts)
      # see if there's "Older posts" button
      url <- current_page %>%
        html_node(".nav-previous a") %>%
        html_attr("href")
    } # end while loop

    tibble::as_tibble(hrbrmstr_posts)
    #> # A tibble: 578 x3
```

We organize the data into the tidy format and append it to our empty
data frame. At the end we will discover that Bob has written over 570
blog articles, which I very much recommend anyone to check out.

## Polite for package developers

If you are developing a package which accesses the web, `polite` can be
used either as a *template*, or as a *backend* for your polite web
session.

### Polite template

Just before its ascension to CRAN, the package acquired new
functionality for helping package developers get started on creating
polite web tools for the users. Any modern package developer is probably
familiar with excellent [`usethis`
package](https://github.com/r-lib/usethis) by Rstudio team. `usethis` is
a collection of scripts for automating package development workflow.
Many `usethis` functions automating repetitive tasks start with prefix
`use_` indicating that what followed will be adopted and “used” by the
package user developes. For details about `use_` family of functions,
see [package
documentation](https://usethis.r-lib.org/reference/index.html).

`{polite}` has one usethis-like function called `polite::use_manners()`.

``` r
polite::use_manners()
```

When called within the analysis (or package) directory, it creates a new
file called `R/polite-scrape.R` (creating `R` directory if necessary)
and populates it with template functions for creating polite
web-scraping session. The functions provided by `polite::use_manners()`
are drop-in replacements for two of the most popular tools in
web-accessing R ecosystem: `read_html()` and `download.file()`. The only
difference is that these functions have `polite_` prefix. In all other
respects they should have look and feel of the original, i.e. in most
cases you should be able to simply replace calls to `read_html()` with
`polite_read_html()` and `download.file` with `polite_download_file()`
and your code should work (provided you scrape from a `url`, which it
the first required argument in both functions).

### Polite backend

Recent addition to polite package is a
[`purrr`-like](https://purrr.tidyverse.org/reference/index.html#section-adverbs)
adverb `politely()` which can make any web-accessing function “polite”
by wrapping it with a code which delivers on four pillars of polite
session:

> **Introduce Yourself, Seek Permission, Take Slowly and Never Ask
> Twice**.

Adverbs can be useful, when a user (package developer) wants to
“delegate” polite session handling to external package, without
modifying the existing code. The only thing user needs to do is wrap
existing verb with `politely()` and use the new function instead of the
original.

Let’s say you wanted to use `httr::GET` for accessing certain API, such
as `musicbrainz` and extract certain data from a deeply nested list,
returned by the server. Your originally developed code looks like this:

``` r
library(magrittr)
#> 
#> Attaching package: 'magrittr'
#> The following object is masked from 'package:purrr':
#> 
#>     set_names
library(httr)
library(xml2)
library(purrr)

beatles_res <- GET("https://musicbrainz.org/ws/2/artist/",
                   query=list(query="Beatles", limit=10),
                   httr::accept("application/json"))
if(!is.null(beatles_res)) beatles_lst <- httr::content(beatles_res, type = "application/json")

str(beatles_lst, max.level = 2)
#> List of 4
#>  $ created: chr "2026-05-10T19:33:09.681Z"
#>  $ count  : int 274
#>  $ offset : int 0
#>  $ artists:List of 10
#>   ..$ :List of 14
#>   ..$ :List of 18
#>   ..$ :List of 17
#>   ..$ :List of 12
#>   ..$ :List of 13
#>   ..$ :List of 18
#>   ..$ :List of 11
#>   ..$ :List of 8
#>   ..$ :List of 5
#>   ..$ :List of 7
```

This code does not comply with `polite` principles. It does not provide
human-readable user-agent string, it does not consult `robots.txt` about
permissions. It is possible to run this code in the loop and
(accidentally) overwhelm the server with requests. It does not cache the
results, so if this code is re-run again, data will be re-queried.

You could write your own infastructure for handling useragent,
robots.txt, rate limiting and memoisation, or you could simply use an
adverb `politely()` which does all of these things for you.

### Querying colormind.io with polite backend

Here’s an example from using colormind.io API. We will need a couple of
service functions to convert colors between HEX and RGB and to prepare a
json [required by the service](http://colormind.io/api-access/).

``` r
rgba2hex <- function(r,g,b,a) {grDevices::rgb(r, g, b, a, maxColorValue = 255)}

hex2rgba <- function(x, alpha=TRUE){t(grDevices::col2rgb(x, alpha = alpha))}

prepare_colormind_query <- function(x, model){
  lst <- list(model=model)

  if(!is.null(x)){
    x <- utils::head(c(x, rep(NA_character_, times=4)), 5) # pad it with NAs
    x_mat <- hex2rgba(x)
    x_lst <- lapply(seq_len(nrow(x_mat)), function(i) if(x_mat[i,4]==0) "N" else x_mat[i,1:3])
    lst <- c(list(input=x_lst), lst)
  }
  jsonlite::toJSON(lst, auto_unbox = TRUE)
}
```

Now all we have to do is to “wrap” existing function in the `politely`
adverb. Then call the new function insted of original. You dont need to
change anything other than a function name.

``` r
polite_GET <- politely(httr::GET, verbose=TRUE)

#res <- httr::GET("http://colormind.io/list") # was
res <- polite_GET("http://colormind.io/list") # now
#> Fetching robots.txt
#> rt_robotstxt_http_getter: normal http get
#> Warning in request_handler_handler(request = request, handler = on_not_found, :
#> Event: on_not_found
#> Warning in request_handler_handler(request = request, handler =
#> on_file_type_mismatch, : Event: on_file_type_mismatch
#> Warning in request_handler_handler(request = request, handler =
#> on_suspect_content, : Event: on_suspect_content
#> 
#> New copy robots.txt was fetched from http://colormind.io/robots.txt
#> Total of 0 crawl delay rule(s) defined for this host.
#> Your rate will be set to 1 request every 5 second(s).
#> Pausing...
#> Scraping: http://colormind.io/list
#> Setting useragent: polite R (4.6.0 x86_64-w64-mingw32 x86_64 mingw32) bot
jsonlite::fromJSON(httr::content(res, as = "text"))$result
#> [1] "ui"                    "default"               "pokemon_gold_silver"  
#> [4] "flower_photography"    "raise_the_red_lantern" "kaguya_film"
```

The backend functionality of `polite` can be used for *any* function as
long as it has `url` argument (or the first argument is a url). Here’s
an example of polite POST created with adverb `politely`.

``` r
polite_POST <- politely(POST, verbose=TRUE)

clue_colors <-c(NA, "lightseagreen", NA, "coral", NA)

req <- prepare_colormind_query(clue_colors, "default")

#res <- httr::POST(url='http://colormind.io/api/', body = req) #was
res <- polite_POST(url='http://colormind.io/api/', body = req) #now
#> Fetching robots.txt
#> rt_robotstxt_http_getter: cached http get
#> Warning in request_handler_handler(request = request, handler = on_not_found, :
#> Event: on_not_found
#> Warning in request_handler_handler(request = request, handler =
#> on_file_type_mismatch, : Event: on_file_type_mismatch
#> Warning in request_handler_handler(request = request, handler =
#> on_suspect_content, : Event: on_suspect_content
#> 
#> Found the cached version of robots.txt for http://colormind.io/robots.txt
#> Total of 0 crawl delay rule(s) defined for this host.
#> Your rate will be set to 1 request every 5 second(s).
#> Pausing...
#> Scraping: http://colormind.io/api/
#> Setting useragent: polite R (4.6.0 x86_64-w64-mingw32 x86_64 mingw32) bot
res_json <- httr::content(res, as = "text")
res_mcol <- jsonlite::fromJSON(res_json)$result
colrs <- rgba2hex(res_mcol)
scales::show_col(colrs, ncol = 5)
```

<img src="man/figures/README-unnamed-chunk-10-1.png" alt="" width="100%" />

### Querying musicbrainz API with polite backend

[Musicbrainz API](https://musicbrainz.org/doc/MusicBrainz_API) allows
querying data on artists, releases, labels and all things music. API
endpoint, unfortunately, is Disallowed in `robots.txt`, but it is
completely legal to access for small size requests. Mass querying is
easier using a datadump, with musicbrainz published periodically. We can
create polite GET and turn off `robots.txt` validation.

``` r
library(polite)
polite_GET_nrt <- politely(GET, verbose=TRUE, robots = FALSE) # turn off robotstxt checking

beatles_lst <- polite_GET_nrt("https://musicbrainz.org/ws/2/artist/",
                   query=list(query="Beatles", limit=10),
                   httr::accept("application/json")) %>%
  httr::content(type = "application/json")
#> Pausing...
#> Scraping: https://musicbrainz.org/ws/2/artist/
#> Setting useragent: polite R (4.6.0 x86_64-w64-mingw32 x86_64 mingw32) bot
str(beatles_lst, max.level = 2)
#> List of 4
#>  $ created: chr "2026-05-10T19:33:09.681Z"
#>  $ count  : int 274
#>  $ offset : int 0
#>  $ artists:List of 10
#>   ..$ :List of 14
#>   ..$ :List of 18
#>   ..$ :List of 17
#>   ..$ :List of 12
#>   ..$ :List of 13
#>   ..$ :List of 18
#>   ..$ :List of 11
#>   ..$ :List of 8
#>   ..$ :List of 5
#>   ..$ :List of 7
```

Lets parse the response

``` r
options(knitr.kable.NA = '')
beatles_lst %>%
  extract2("artists") %>%
  {tibble::tibble(id=map_chr(.,"id", .default=NA_character_),
                  match_pct=map_int(.,"score", .default=NA_character_),
                  type=map_chr(.,"type", .default=NA_character_),
                  name=map_chr(., "name", .default=NA_character_),
                  country=map_chr(., "country", .default=NA_character_),
                  lifespan_begin=map_chr(., c("life-span", "begin"),.default=NA_character_),
                  lifespan_end=map_chr(., c("life-span", "end"),.default=NA_character_)
                  )
    } %>% knitr::kable(col.names = c(id="Musicbrainz ID", match_pct="Match, %",
                                     type="Type", name="Name of artist",
                                     country="Country", lifespan_begin="Career begun",
                                     lifespan_end="Career ended"))
```

| Musicbrainz ID | Match, % | Type | Name of artist | Country | Career begun | Career ended |
|:---|---:|:---|:---|:---|:---|:---|
| b10bbbfc-cf9e-42e0-be17-e2c3e1d2600d | 100 | Group | The Beatles | GB | 1960 | 1970-04-10 |
| 4d5447d7-c61c-4120-ba1b-d7f471d385b9 | 72 | Person | John Lennon | GB | 1940-10-09 | 1980-12-08 |
| ba550d0e-adac-4864-b88b-407cab5e76af | 72 | Person | Paul McCartney | GB | 1942-06-18 |  |
| 5e685f9e-83bb-423c-acfa-487e34f15ffd | 69 | Group | The Tape-beatles | US | 1986-12 |  |
| 0f697bc6-6df7-41c9-b550-39981e520d70 | 68 | Group | The Beatles Revival Band | DE | 1976 |  |
| 42a8f507-8412-4611-854f-926571049fa0 | 67 | Person | George Harrison | GB | 1943-02-24 | 2001-11-29 |
| 1019b551-eba7-4e7c-bc7d-eb427ef54df2 | 64 | Group | Blues Beatles | BR | 2010 |  |
| 9d953ee6-4ea6-4b0e-aea6-7268d380bef1 | 64 | Group | The Beatles |  |  |  |
| 3f944209-cab5-4c54-98ff-7efd34463434 | 63 |  | D-Beatles |  |  |  |
| 5a45e8c5-e8e5-4f05-9429-6dd00f0ab50b | 63 | Group | Instrumental Beatles |  |  |  |

## Learn more

[Ethical webscraper
manifesto](https://medium.com/data-science/ethics-in-web-scraping-b96b18136f01)

Package logo uses elements of a free image by
[pngtree.com](https://pngtree.com)

\[1\] Wiktionary (2018), The free dictionary, retrieved from
<https://en.wiktionary.org/wiki/bow_and_scrape>

#' null-coalescing operator. See purr for details.
#'
#' @param lhs legt hand side
#' @param rhs right hand side
#' @examples
#' a %otherwise% b
#'
`%otherwise%` <- function(lhs, rhs) {
  if (!is.null(lhs) && length(lhs) > 0) lhs else rhs
}

#' Function to get robots.txt is structured form. Memoised
#'
#' @param ... arguments passed to `robotstxt::robotstxt()`
#' @param user_agent user agent string
#' @param delay default delay
#' @param verbose logical
#' @examples
polite_fetch_rtxt <- memoise::memoise(function(..., user_agent, delay, verbose){
  rt <- robotstxt::robotstxt(...)
  delay_df <- rt$crawl_delay
  crawldelays <-   as.numeric(delay_df[with(delay_df, useragent==user_agent), "value"]) %otherwise%
    as.numeric(delay_df[with(delay_df, useragent=="*"), "value"]) %otherwise% 0

  rt$delay_rate <- max(crawldelays, delay, 1)

  if(verbose){
    message("Bowing to: ", rt$domain)
    message("There's ", nrow(delay_df), " crawl delay rule(s) defined for this host.")
    message("Your rate will be set to 1 request every ", rt$delay_rate, " second(s).")}

  rt
})

#' Function for checking robots.txt file
#'
#' @param url web address for download
#' @param delay default delay
#' @param user_agent user agent string
#' @param force force re-downloading of robots.xtx
#' @param verbose logical
#'
#' @return
#'
#' @examples
check_rtxt <-function(url, delay, user_agent, force, verbose){
  url_parsed <- httr::parse_url(url)
  host_url <- paste0(url_parsed$scheme, "://", url_parsed$hostname)
  rt <- polite_fetch_rtxt(host_url, force=force, user_agent=user_agent, delay=delay, verbose=verbose)
  is_scrapable <- rt$check(paths=url_parsed$path, bot=user_agent)

  if(is_scrapable)
    Sys.sleep(rt$delay_rate)
  else
    warning("robots.txt says this path is NOT scrapable for your user agent!", call. = FALSE)

  is_scrapable
}

#' function that actually fetches response from the web
#'
#' @param url web address for scraping
#' @param ... arguments passed to `httr::GET()`
#' @param delay scrapting delay. Default 5 sec
#' @param user_agent user agent string. Default value `paste0("polite ", getOption("HTTPUserAgent"), "bot")`
#' @param force force re-download of robots.txt
#' @param verbose default FALSE
#'
#' @return
#'
#' @examples
polite_read_html <- memoise::memoise(
                   function(url, ...,
                   delay = 5,
                   user_agent=paste0("polite ", getOption("HTTPUserAgent"), "bot"),
                   force = FALSE,
                   verbose=FALSE){

  if(!check_rtxt(url, delay, user_agent, force, verbose)){
    return(NULL)
  }
# this is not working yet.
#  old_ua <-  getOption("HTTPUserAgent")
#  options("HTTPUserAgent"= user_agent)
  if(verbose) message("Scraping: ", url)
  res <- httr::GET(url, user_agent(user_agent), ...)
#  options("HTTPUserAgent"= old_ua)
  httr::content(res)
})


#' Guess filename for download from url
#'
#' @param x url to guess filename from
#'
#' @return
#'
#' @examples
guess_basename <- function(x) {
  destfile <- basename(x)
  if(tools::file_ext(destfile)==""){
    hh <- httr::HEAD(x)
    destfile <- basename(hh$url)
    if(tools::file_ext(destfile)==""){
      cds <- httr::headers(hh)$`content-disposition`
      destfile <- gsub('.*filename=', '', gsub('\\\"','', cds))
    }}
  destfile %otherwise% basename(x)
}


#' Polite download
#'
#' @param url web address for the file to be downloaded
#' @param destfile name of destination file
#' @param ... additional arguments passed to `download.file`
#' @param quiet default value is inverse of  `verbose`
#' @param mode download mode. Default value is "wb"
#' @param path path to save. Default path `downloads/`
#' @param user_agent default value `paste0("polite ", getOption("HTTPUserAgent"))`
#' @param delay default value equal 5
#' @param force force re-download of robots.txt
#' @param overwrite overwrite downloaded file. Default value FALSE
#' @param verbose default value is FALSE
#'
#' @return
#'
#' @examples
polite_download_file <- memoise::memoise(
                        function(url, destfile=guess_basename(url), ...,
                             quiet=!verbose, mode = "wb", path="downloads/",
                             user_agent=paste0("polite ", getOption("HTTPUserAgent")),
                             delay = 5, force = FALSE, overwrite=FALSE, verbose=FALSE){

  if(!check_rtxt(url, delay, user_agent, force, verbose)) return(NULL)

  if(!dir.exists(path)) dir.create(path)

  destfile <- paste0(path, destfile)

  if(file.exists(destfile) && !overwrite){
    message("File ", destfile, " already exists!")
    return(destfile)
  }

  old_ua <-  getOption("HTTPUserAgent")
  on.exit(options("HTTPUserAgent"= old_ua), add = TRUE)
  options("HTTPUserAgent"= user_agent)
  if(verbose) message("Scraping: ", url)
  utils::download.file(url=url, destfile=destfile, quiet=quiet, mode=mode, ...)
  options("HTTPUserAgent"= old_ua)
  destfile
})

#' Introduce yourself to the host
#'
#' @param url URL
#' @param user_agent character value passed to user agent string
#' @param delay desired delay between scraping attempts. Final value will be the maximum of desired and mandated delay, as stipulated by `robots.txt` for relevant user agent
#' @param times number of times to attempt scraping. Default is 3.
#' @param force refresh all memoised functions. Clears up `robotstxt` and `scrape` caches. Default is `FALSE`
#' @param verbose TRUE/FALSE
#' @param ... other curl parameters wrapped into `httr::config` function
#'
#' @return object of class `polite`, `session`
#'
#' @examples
#' \donttest{
#'  library(polite)
#'
#'  host <- "https://www.cheese.com"
#'  session <- bow(host)
#'  session
#' }
#' @rdname bow
#' @importFrom robotstxt robotstxt
#' @importFrom httr parse_url handle config add_headers GET
#' @importFrom ratelimitr limit_rate rate get_rates
#' @importFrom memoise forget
#' @importFrom stats na.omit
#' @importFrom utils download.file
#' @export
bow <- function(url,
                user_agent = "polite R package",
                delay = 5,
                times = 3,
                force = FALSE, verbose=FALSE,
                ...){

  stopifnot("Character user agent is required"=is.character(user_agent))
  stopifnot("Single user agent is required. Please provide character vector of length 1"=(length(user_agent) == 1)) # write meaningful error ref Lionel talk
  stopifnot("Character URL is required"=is.character(url))
  stopifnot("Single URL should be provided. Please provide character vector of length 1"=(length(url) == 1)) # write meaningful error ref Lionel talk
  stopifnot("Number of times to attempt scraping should be numeric"=is.numeric(times))
  stopifnot("Number of times to attempt scraping should be positive"=(times>0))
  if(force) memoise::forget(scrape)

  url_parsed <- httr::parse_url(url)
  url_subdomain <- paste0(url_parsed$scheme, "://", url_parsed$hostname)
  rt <- robotstxt::robotstxt(domain = url_subdomain,
                            user_agent = user_agent,
                            warn = verbose, force = force)

  delay_df <- rt$crawl_delay
  delay_rt <- as.numeric(delay_df[with(delay_df, useragent==user_agent), "value"]) %otherwise%
    as.numeric(delay_df[with(delay_df, useragent=="*"), "value"]) %otherwise% 0

  # define object
  self <- structure(
    list(
      handle   = httr::handle(url),
      config   = c(httr::config(autoreferer = 1L),
                   httr::add_headers("user-agent"=user_agent),...),
      url      = url,
      back     = character(),
      forward  = character(),
      response = NULL,
      html     = new.env(parent = emptyenv(), hash = FALSE),
      user_agent = user_agent,
      domain   =  url_subdomain,
      robotstxt= rt,
      delay  = max(delay_rt, delay),
      times = times
    ),
    class = c("polite", "session")
  )

  if(verbose && !is_scrapable(self))
    warning("Psst!...It's not a good idea to scrape here!", call. = FALSE)

  if(self$delay<5)
    if(grepl("polite|dmi3kno", self$user_agent)){
      stop("You cannot scrape this fast. Please reconsider delay period.", call. = FALSE)
    warning("This is a little too fast. Are you sure you want to risk being banned?", call. = FALSE)
    }

  # set new rate limits
  if(self$delay != ratelimitr::get_rates(httr_get_ltd)[[1]]["period"]){
    set_scrape_delay(self$delay)
  }

  if(self$delay != ratelimitr::get_rates(download_file_ltd)[[1]]["period"]){
    set_rip_delay(self$delay)
  }

  self
}


#' Reset scraping/ripping rate limit
#'
#' @param delay Delay between subsequent requests. Default for package is 5 sec.
#' It can be set lower only under the condition of specifying a custom user-agent string.
#'
#' @return Updates rate-limit property of `scrape` and `rip` functions, respectively.
#'
#' @examples
#' \donttest{
#'  library(polite)
#'
#'  host <- "https://www.cheese.com"
#'  session <- bow(host)
#'  session
#' }
#' @rdname set_delay
#' @importFrom ratelimitr UPDATE_RATE rate
#' @export
set_scrape_delay <- function(delay){
  ratelimitr::UPDATE_RATE(httr_get_ltd,ratelimitr::rate(n=1, period = delay))
}

#' @rdname set_delay
#' @importFrom ratelimitr UPDATE_RATE rate
#' @export
set_rip_delay <- function(delay){
  ratelimitr::UPDATE_RATE(download_file_ltd,ratelimitr::rate(n=1, period = delay))
}

#' Print host introduction object
#'
#' @param x object of class `polite`, `session`
#' @param ... other parameters passed to methods
#' @export
print.polite <- function(x, ...) {
  cat(paste0("<polite session> ", x$url, "\n",
             "    ", "User-agent: ", x$user_agent, "\n",
             "    ", "robots.txt: ", nrow(x$robotstxt$permissions), " rules are defined for ",length(x$robotstxt$bots), " bots\n",
             "   ", "Crawl delay: ", x$delay," sec\n"))
  if(is_scrapable(x)){
    cat(" ", "The path is scrapable for this user-agent\n")
  } else {
    cat(" ", "The path is not scrapable for this user-agent\n")
  }
}

#' @param x object of class `polite`, `session`
#' @rdname bow
#' @export
is.polite <- function(x) inherits(x, "polite")

context("Bow")
library(polite)
library(webmockr)

webmockr::enable(adapter = "httr")

webmockr::httr_mock()

webmockr::stub_registry_clear()

rb_txt <- "User-agent: *\nDisallow: /deny\n"

#### mocking robotstxt request

webmockr::stub_request("get", "https://httpbin.org/robots.txt") %>%
  webmockr::wi_th(
    headers = list('Accept' = 'text/html',
                   'user-agent' = 'polite R package')) %>%
  webmockr::to_return(body=rb_txt, status=200)

test_that("Argument assurances in bow work", {
  expect_error(bow("https://httpbin.org/get", user_agent = c("polite", "R", "package")), regexp = "length")
  expect_error(bow("https://httpbin.org/get", user_agent =42L), regexp = "Character")
  expect_error(bow(url=42L, user_agent ="polite R package"), regexp = "Character")
  expect_error(bow(url = c("https://httpbin.org/get", "https://httpbin.org/post")), regexp = "length")
  expect_error(bow("https://httpbin.org/get", times=0), regexp = "times")
  expect_error(bow("https://httpbin.org/get", times="five"), regexp = "times")
  })

webmockr::stub_registry_clear()
webmockr::disable(adapter = "httr")

context("Scrape")
library(webmockr)
library(testthat)

webmockr::enable(adapter = "httr")

webmockr::httr_mock()

webmockr::stub_registry_clear()

#### stub robotstxt request

rb_txt <- "User-agent: *\nDisallow: /deny\n"

webmockr::stub_request("get", "https://httpbin.org/robots.txt") %>%
  wi_th(
    headers = list('Accept' = 'application/json, text/xml, application/xml, */*',
                   'user-agent' = 'polite R package')) %>%
  webmockr::to_return(body=rb_txt, status=200)

#### stub scraping request
rq_txt <- list()

for (i in 1:2){
 rq_txt[i]<- paste0('{"args": {"q": ',
                  i,'},"headers": {"Accept": "application/json","User-Agent": "polite R package"}',
                  ',"url": "https://httpbin.org/get?q=',i,'"}')

 stub_request("get", paste0("https://httpbin.org/get?q=",i)) %>%
   wi_th(
     headers = list('Accept' = 'application/json',
                    'user-agent' = 'polite R package')) %>%
   webmockr::to_return(body=rq_txt[[i]], status=200)
}
#webmockr::stub_registry()

base_url <- "https://httpbin.org/get"
query_lst <- list(q=1)
session <- polite::bow(base_url)
res <- scrape(session, query = query_lst, accept = "application/json", verbose = TRUE)

test_that("bow type is controlled", {
  expect_error(scrape(base_url), regexp = "bow")
})

test_that("query is controlled", {
  expect_error(scrape(session, query = c(q=1)), regexp = "is.list")
})

test_that("scrape is returning proper result",{
  expect_equal(res, charToRaw(rq_txt[[1]]))
})

session <- polite::bow(base_url, force = TRUE)

test_that("Forced bow is un-memoised", {
  expect_false(memoise::has_cache(polite::scrape)(bow=session))
})

webmockr::disable(adapter = "httr")

#' @importFrom httr http_error content modify_url parse_url
m_scrape <- function(
  bow,
  query = NULL,
  params = NULL,
  accept = "html",
  content = NULL,
  verbose = FALSE
) {
  # nolint

  if (!inherits(bow, "polite")) {
    stop("Please be polite: bow then scrape!")
  }

  stopifnot(is.list(query) || is.null(query))

  if (!is.null(query)) {
    bow$url <- httr::modify_url(bow$url, query = query)
  } else {
    if (!is.null(params)) {
      warning(
        "Argument `params` is deprecated. Please, use `query` passing parameters as a named list.",
        call. = FALSE
      )
      params <- gsub("^.*\\?", "", params) # this will delete url prefix if any
      query <- httr::parse_url(paste0(
        "http://www.example.com/q?",
        params
      ))$query
      bow$url <- httr::modify_url(bow$url, query = query)
    }
  }

  if (!is_scrapable(bow)) {
    warning("No scraping allowed here!", call. = FALSE)
    return(NULL)
  }

  if (substr(accept, 1, 1) != "." && !grepl("/", accept)) {
    accept <- paste0("example.", accept)
  }

  accept_type <- httr::accept(accept)
  bow$config <- c(bow$config, accept_type)

  response <- httr_get_ltd(bow$url, bow$config, bow$handle, bow$times, verbose)

  if (httr::http_error(response)) {
    warning(httr::http_status(response)$message, " ", bow$url, call. = FALSE)
    return(NULL)
  }

  content <- content %otherwise% response$headers$`content-type`

  res <- tryCatch(
    {
      httr::content(response, type = content)
    },
    error = function(cond) {
      warning(
        paste0(
          "<polite session> Encountered an error while parsing content.\n",
          "There seems to be a mismatch of content type or encoding or both.\n",
          "The server says it is serving:",
          response$headers$`content-type`,
          "\n",
          "Here's the text of the error generated by httr::content(): \n",
          cond,
          "\n",
          "Returning raw vector which can be parsed with rawToChar(). Good luck!\n"
        ),
        call. = FALSE
      )
      return(httr::content(response, as = "raw"))
    }
  )
  res
}


#' Scrape the content of authorized page/API
#'
#' @param bow host introduction object of class `polite`, `session` created by `bow()` or `nod()`
#' @param query named list of parameters to be appended to URL in the format `list(param1=valA, param2=valB)`
#' @param params deprecated. Use `query` argument above.
#' @param accept character value of expected data type to be returned by host (e.g. `html`, `json`, `xml`, `csv`, `txt`, etc.)
#' @param content MIME type (aka internet media type) used to override the content type returned by the server.
#' See http://en.wikipedia.org/wiki/Internet_media_type for a list of common types. You can add the `charset` parameter to override the server's default encoding
#' For example, you could write `content="text/html; charset=UTF-8"`
#' @param verbose extra feedback from the function. Defaults to `FALSE`
#'
#' @return Object of class `httr::response` which can be further processed by functions in `rvest` package
#'
#' @examples
#' \donttest{
#'  library(rvest)
#'   bow("https://en.wikipedia.org/wiki/List_of_cognitive_biases") %>%
#'    scrape(content="text/html; charset=UTF-8") %>%
#'    html_nodes(".wikitable") %>%
#'    html_table()
#'}
#'
#' @export
scrape <- memoise::memoise(m_scrape)

#' Agree modification of session path with the host
#'
#' @param bow object of class `polite`, `session` created by `polite::bow()`
#' @param path string value of path/URL to follow. The function accepts either a path (string part of URL following domain name) or a full URL
#' @param verbose `TRUE`/`FALSE`
#' @return object of class `polite`, `session` with modified URL
#'
#' @examples
#' \donttest{
#'  library(polite)
#'
#'  host <- "https://www.cheese.com"
#'  session <- bow(host) %>%
#'               nod(path="by_type")
#'  session
#' }
#' @importFrom httr parse_url modify_url
#' @export
nod <- function(bow, path, verbose=FALSE){

  if(!inherits(bow, "polite"))
    stop("Please bow before you nod", call. = FALSE)

  # if user supplied URL instead of path
  if(grepl("://|www\\.", path)){
    url_parsed <- httr::parse_url(path)
    url_subdomain <- paste0(url_parsed$scheme, "://", url_parsed$hostname)
    if(url_subdomain!=bow$domain)
      bow <- bow(url = path, user_agent = bow$user_agent, delay=bow$delay, times=bow$times,
                 force=FALSE, verbose=verbose, bow$config)
    path <- url_parsed$path
  }

  bow$url <- httr::modify_url(bow$url, path=path)

  if(verbose && !is_scrapable(bow))
    warning("Psst!...It's not a good idea to scrape here!", call. = FALSE)

  bow
}

#' Polite file download
#'
#' @param bow host introduction object of class `polite`, `session` created by `bow()` or `nod()`
#' @param destfile optional new file name to use when saving the file. If missing, it will be guessed from `basename(url)``
#' @param ... other parameters passed to `download.file`
#' @param mode character. The mode with which to write the file. Useful values are `w`, `wb` (binary), `a` (append) and `ab`. Not used for methods `wget` and `curl`.
#' @param path character. Path where to save the destfile. By default is temporary directory created with `tempdir()`
#' Ignored if `destfile` contains path along with filename.
#' @param overwrite if `TRUE` will overwrite file on disk
#'
#' @return Full path to the locally saved file indicated by the user in `destfile` (and `path`)
#' @export
#'
#' @examples
#' \donttest{
#' bow("https://en.wikipedia.org/") %>%
#'  nod("wiki/Flag_of_the_United_States#/media/File:Flag_of_the_United_States.svg") %>%
#'  rip()
#' }
rip <- function(bow, destfile=NULL, ..., mode="wb", path=tempdir(), overwrite=FALSE){

  url <- bow$url

  if (is.null(destfile)) destfile <- basename(url)

  if(dirname(destfile)!=dirname(".")) path <- dirname(destfile)

  if(!dir.exists(path)) dir.create(path, recursive = TRUE)

  destfilepath <- file.path(path, destfile)

  if(file.exists(destfilepath) && !overwrite){
    warning("File already exists", call. = FALSE)
    return(destfilepath)
    }

  download_file_ltd(url, destfilepath, mode=mode, ...)

  return(destfilepath)
}

#' Null coalescing operator
#'
#'
#' @name null-coalesce
#' @rdname nullcoalesce
#' @keywords internal
#' @usage lhs \%otherwise\% rhs
#'
`%otherwise%` <- function(lhs, rhs) {
  if (!is.null(lhs) && length(lhs) > 0) lhs else rhs
}

#' Pipe operator
#'
#' See \code{magrittr::\link[magrittr]{\%>\%}} for details.
#'
#' @name %>%
#' @rdname pipe
#' @keywords internal
#' @export
#' @importFrom magrittr %>%
#' @usage lhs \%>\% rhs
NULL

is_scrapable <- function(bow){
  is_scrapable_rt(bow$robotstxt, bow$url, bow$user_agent)
}

#' @importFrom httr parse_url
is_scrapable_rt <- function(rtxt, url, user_agent){
  url_parsed <- httr::parse_url(url)
  rtxt$check(paths=url_parsed$path, bot=user_agent)
}

#' @importFrom httr GET
httr_get <- function(url, config, handle, times, verbose){
  httr::RETRY(verb="GET",
    url = url,
    config = config,
    handle = handle,
    times = times,
    pause_base = 5,
    quiet=!verbose
  )
}

#' @importFrom ratelimitr limit_rate rate
httr_get_ltd <- ratelimitr::limit_rate(httr_get,
                                       ratelimitr::rate(n = 1, period=5))

#' @importFrom ratelimitr limit_rate rate
#' @importFrom utils download.file
download_file_ltd <- ratelimitr::limit_rate(utils::download.file,
                                       ratelimitr::rate(n = 1, period=5))



#' Guess download file name from the URL
#'
#' @param x url to guess basename from
#'
#' @return guessed file name
#' @export
#'
#' @examples
#' guess_basename("https://bit.ly/polite_sticker")
#'
#' @importFrom tools file_ext
#' @importFrom httr HEAD headers
guess_basename <- function(x) {
  destfile <- basename(x)
  if(tools::file_ext(destfile)==""){
    hh <- httr::HEAD(x)
    destfile <- basename(hh$url)
    if(tools::file_ext(destfile)==""){
    cds <- httr::headers(hh)$`content-disposition`
    destfile <- gsub('.*filename=', '', gsub('\\\"','', cds))
  }}
  destfile %otherwise% basename(x)
}


is_url <- function(x){
  # from Jim Hester's rex vignette https://cran.r-project.org/web/packages/rex/vignettes/url_parsing.html
  re <- "^(?:(?:http(?:s)?|ftp)://)(?:\\S+(?::(?:\\S)*)?@)?(?:(?:[a-z0-9\u00a1-\uffff](?:-)*)*(?:[a-z0-9\u00a1-\uffff])+)(?:\\.(?:[a-z0-9\u00a1-\uffff](?:-)*)*(?:[a-z0-9\u00a1-\uffff])+)*(?:\\.(?:[a-z0-9\u00a1-\uffff]){2,})(?::(?:\\d){2,5})?(?:/(?:\\S)*)?$"
  grepl(re, x)
}

#' Convert collection of html nodes into data frame
#'
#' @param x `xml_nodeset` object, containing text and attributes of interest
#' @param attrs character vector of attribute names. If missing, all attributes will be used
#' @param trim if `TRUE`, will trim leading and trailing spaces
#' @param defaults character vector of default values to be passed to `rvest::html_attr()`. Recycled to match length of `attrs`
#' @param add_text if `TRUE`, node content will be added as `.text` column (using `rvest::html_text`)
#'
#' @return data frame with one row per xml node, consisting of an html_text column with text and additional columns with attributes
#' @export
#'
#' @examples
#' \donttest{
#' library(polite)
#' library(rvest)
#' bow("https://en.wikipedia.org/wiki/List_of_cognitive_biases") %>%
#'   scrape() %>%
#'   html_nodes("tr td:nth-child(1) a") %>%
#'   html_attrs_dfr()
#'   }
#' @importFrom rvest html_text html_attrs html_attr
#' @importFrom stats setNames
html_attrs_dfr <- function(x, attrs=NULL, trim=FALSE, defaults=NA_character_, add_text=TRUE){

  attrs <-  attrs %otherwise% unique(unlist(lapply(rvest::html_attrs(x), names)))
  defaults <- rep(defaults, length.out=length(attrs))
  attrs_lst <- lapply(seq_along(attrs), function(i)
                               rvest::html_attr(x, attrs[i], default = defaults[i]))
  if(trim) attrs_lst <- lapply(attrs_lst, trimws)
  attrs_lst <- stats::setNames(attrs_lst, attrs)
  res <- as.data.frame(do.call(cbind, attrs_lst), stringsAsFactors=FALSE, row.names = FALSE)

  if(add_text) res$.text <- rvest::html_text(x, trim=trim)
  res
}

extract_domain <- function(url){
  url_parsed <- httr::parse_url(url)
  paste0(url_parsed$scheme, "://", url_parsed$hostname)
}


fetch_rtxt <-function(domain, user_agent, delay, force, verbose){
  rt_txt <- robotstxt::get_robotstxt(domain=domain, force=force, user_agent = user_agent, verbose = verbose)
  rq <- attr(rt_txt, "request")
  rt <- robotstxt::robotstxt(text=rt_txt)
  rt_delay_df <- rt$crawl_delay
  crawldelay <-   as.numeric(with(rt_delay_df, value[useragent==user_agent])) %otherwise%
    as.numeric(with(rt_delay_df, value[useragent=="*"])) %otherwise% 0

  rt$delay_rate <- max(crawldelay, delay, 1)
  rt$received_from <- rq$url
  rt$cached <- attr(rt_txt, "cached")

  if(verbose){
    if(rt$cached)
      message("\nFound the cached version of robots.txt for ", rt$received_from)
      else
      message("\nNew copy robots.txt was fetched from ", rt$received_from)
    message("Total of ", nrow(rt_delay_df), " crawl delay rule(s) defined for this host.")
    message("Your rate will be set to 1 request every ", rt$delay_rate, " second(s).")}

  rt
}


#' Give your web-scraping function good manners polite
#'
#' @param fun function to be turned "polite". Must contain an argument named `url`, which contains url to be queried.
#' @param user_agent optional, user agent string to be used. Defaults to `paste("polite", getOption("HTTPUserAgent"), "bot")`
#' @param robots optional, should robots.txt be consulted for permissions. Default is TRUE
#' @param force whether or not tp force fresh download of robots.txt
#' @param delay minimum delay in seconds, not less than 1. Default is 5.
#' @param verbose output more information about querying process
#' @param cache memoise cache function for storing results. Default `memoise::cache_memory()`
#'
#' @return polite function
#' @export
#'
#' @examples
#'
#' polite_GET <- politely(httr::GET)
#' @importFrom memoise memoise has_cache cache_memory
politely <- function(fun, user_agent=paste0("polite ", getOption("HTTPUserAgent"), " bot"),
                     robots=TRUE, force=FALSE, delay=5, verbose=FALSE, cache=memoise::cache_memory()){
  f_formals <- formals(args(fun))
  mem_fun <- memoise::memoise(fun, cache=cache)

  if(!"url" %in% names(f_formals) && verbose)
    message("It does not look like there's an argument with the name 'url' in this function. polite::politely() will assume that first argument is a url.")

  function(...){
    arg_lst <- list(...)
    af <- match_to_formals(arg_lst, f_formals)
    if("url" %in% names(af) && is_url(af[["url"]])){
      url <- af[["url"]]
    } else {
      if(is_url(af[[1]])){
        url <- af[[1]]
      } else {
        stop("I can't find an argument containing url. Aborting.", call. = FALSE)
        return(NULL)
      }
    }

    if(robots && !memoise::has_cache(mem_fun)(...)){
      if(verbose) message("Fetching robots.txt")
            hst <- extract_domain(url)
      rtxt <- fetch_rtxt(domain=hst, user_agent = user_agent, delay=delay, force=force, verbose = verbose)
      delay <- rtxt$delay_rate

      if(!is_scrapable_rt(rtxt, url, user_agent)){
        warning("Unfortunately, robots.txt indicates that this path is NOT scrapable for your user agent.", call. = FALSE)
        return(NULL)
      }
    }

    if(!memoise::has_cache(mem_fun)(...)){
      if(verbose) message("Pausing... ")
      Sys.sleep(delay)
      if(verbose) message("Scraping: ", url)
    } else {
      if (verbose) message("Cached results are found! Retrieving results for ", url)
    }

    if(verbose) message("Setting useragent: ", user_agent)
    old_ua <- getOption("HTTPUserAgent")
    on.exit(options("HTTPUserAgent"= old_ua), add = TRUE)
    options("HTTPUserAgent"= user_agent)
    res <- mem_fun(...)
    options("HTTPUserAgent"= old_ua)
    res
  }
}

#' Use manners in your own package or script
#'
#' Creates collection of  `polite` functions for scraping and downloading
#'
#' @param save_as File where function should be created Defaults to "`R/polite-scrape.R`"
#' @param open if `TRUE`, open the resultant files
#' @importFrom usethis use_template
#' @export
use_manners <- function(save_as="R/polite-scrape.R", open = TRUE) {

  save_as_directory <- dirname(save_as)
  if (! dir.exists(save_as_directory)) {
    dir.create(save_as_directory)
  }
  usethis::use_template("polite_template.R", save_as = save_as,
                      open = open, package = "polite")
  invisible()
}

is.nameempty <- function(x){
  nx <- names(x)
  if(is.null(nx))
    return(rep(TRUE, length(x)))
  is.na(nx) | nx==""
}

match_to_formals <- function(a, f, keep_passed_names=TRUE){
  l_wo_names <- is.nameempty(a)
  # named arguments only
  a_w_names <- names(a)[!l_wo_names]
  # how many arguments are unnamed
  l_unn <- length(a[l_wo_names])
  # formals, which have not been explicitly referred to in arguments
  nms_f <- names(f)[!names(f) %in% a_w_names]
  # position of dots in unreferred formals, if not found, assumed to be at the end
  dots_pos <- match("...", nms_f, nomatch = length(nms_f)+1)
  # foreign names (names passed to dots) can be suppressed, but it is probably not a good idea
  if(!keep_passed_names){
    a_w_names <- ifelse(a_w_names %in% names(f), a_w_names, "...")
    names(a)[!l_wo_names] <- a_w_names # this part has names so it is matched literally
  }

  names(a)[l_wo_names] <- c(nms_f[seq_len(min(l_unn, dots_pos-1))],             # this part is matched by position
                            rep.int("...", times=max(0,l_unn-dots_pos+1)))  # this part is passed to dots
  a
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rip.R
\name{rip}
\alias{rip}
\title{Polite file download}
\usage{
rip(
  bow,
  destfile = NULL,
  ...,
  mode = "wb",
  path = tempdir(),
  overwrite = FALSE
)
}
\arguments{
\item{bow}{host introduction object of class \code{polite}, \code{session} created by \code{bow()} or \code{nod()}}

\item{destfile}{optional new file name to use when saving the file. If missing, it will be guessed from `basename(url)``}

\item{...}{other parameters passed to \code{download.file}}

\item{mode}{character. The mode with which to write the file. Useful values are \code{w}, \code{wb} (binary), \code{a} (append) and \code{ab}. Not used for methods \code{wget} and \code{curl}.}

\item{path}{character. Path where to save the destfile. By default is temporary directory created with \code{tempdir()}
Ignored if \code{destfile} contains path along with filename.}

\item{overwrite}{if \code{TRUE} will overwrite file on disk}
}
\value{
Full path to the locally saved file indicated by the user in \code{destfile} (and \code{path})
}
\description{
Polite file download
}
\examples{
\donttest{
bow("https://en.wikipedia.org/") \%>\%
 nod("wiki/Flag_of_the_United_States#/media/File:Flag_of_the_United_States.svg") \%>\%
 rip()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bow.R
\name{bow}
\alias{bow}
\alias{is.polite}
\title{Introduce yourself to the host}
\usage{
bow(
  url,
  user_agent = "polite R package",
  delay = 5,
  times = 3,
  force = FALSE,
  verbose = FALSE,
  ...
)

is.polite(x)
}
\arguments{
\item{url}{URL}

\item{user_agent}{character value passed to user agent string}

\item{delay}{desired delay between scraping attempts. Final value will be the maximum of desired and mandated delay, as stipulated by \code{robots.txt} for relevant user agent}

\item{times}{number of times to attempt scraping. Default is 3.}

\item{force}{refresh all memoised functions. Clears up \code{robotstxt} and \code{scrape} caches. Default is \code{FALSE}}

\item{verbose}{TRUE/FALSE}

\item{...}{other curl parameters wrapped into \code{httr::config} function}

\item{x}{object of class \code{polite}, \code{session}}
}
\value{
object of class \code{polite}, \code{session}
}
\description{
Introduce yourself to the host
}
\examples{
\donttest{
 library(polite)

 host <- "https://www.cheese.com"
 session <- bow(host)
 session
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/use-manners.R
\name{use_manners}
\alias{use_manners}
\title{Use manners in your own package or script}
\usage{
use_manners(save_as = "R/polite-scrape.R", open = TRUE)
}
\arguments{
\item{save_as}{File where function should be created Defaults to "\code{R/polite-scrape.R}"}

\item{open}{if \code{TRUE}, open the resultant files}
}
\description{
Creates collection of  \code{polite} functions for scraping and downloading
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nod.R
\name{nod}
\alias{nod}
\title{Agree modification of session path with the host}
\usage{
nod(bow, path, verbose = FALSE)
}
\arguments{
\item{bow}{object of class \code{polite}, \code{session} created by \code{polite::bow()}}

\item{path}{string value of path/URL to follow. The function accepts either a path (string part of URL following domain name) or a full URL}

\item{verbose}{\code{TRUE}/\code{FALSE}}
}
\value{
object of class \code{polite}, \code{session} with modified URL
}
\description{
Agree modification of session path with the host
}
\examples{
\donttest{
 library(polite)

 host <- "https://www.cheese.com"
 session <- bow(host) \%>\%
              nod(path="by_type")
 session
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bow.R
\name{print.polite}
\alias{print.polite}
\title{Print host introduction object}
\usage{
\method{print}{polite}(x, ...)
}
\arguments{
\item{x}{object of class \code{polite}, \code{session}}

\item{...}{other parameters passed to methods}
}
\description{
Print host introduction object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bow.R
\name{set_scrape_delay}
\alias{set_scrape_delay}
\alias{set_rip_delay}
\title{Reset scraping/ripping rate limit}
\usage{
set_scrape_delay(delay)

set_rip_delay(delay)
}
\arguments{
\item{delay}{Delay between subsequent requests. Default for package is 5 sec.
It can be set lower only under the condition of specifying a custom user-agent string.}
}
\value{
Updates rate-limit property of \code{scrape} and \code{rip} functions, respectively.
}
\description{
Reset scraping/ripping rate limit
}
\examples{
\donttest{
 library(polite)

 host <- "https://www.cheese.com"
 session <- bow(host)
 session
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/politely.R
\name{politely}
\alias{politely}
\title{Give your web-scraping function good manners polite}
\usage{
politely(
  fun,
  user_agent = paste0("polite ", getOption("HTTPUserAgent"), " bot"),
  robots = TRUE,
  force = FALSE,
  delay = 5,
  verbose = FALSE,
  cache = memoise::cache_memory()
)
}
\arguments{
\item{fun}{function to be turned "polite". Must contain an argument named \code{url}, which contains url to be queried.}

\item{user_agent}{optional, user agent string to be used. Defaults to \code{paste("polite", getOption("HTTPUserAgent"), "bot")}}

\item{robots}{optional, should robots.txt be consulted for permissions. Default is TRUE}

\item{force}{whether or not tp force fresh download of robots.txt}

\item{delay}{minimum delay in seconds, not less than 1. Default is 5.}

\item{verbose}{output more information about querying process}

\item{cache}{memoise cache function for storing results. Default \code{memoise::cache_memory()}}
}
\value{
polite function
}
\description{
Give your web-scraping function good manners polite
}
\examples{

polite_GET <- politely(httr::GET)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scrape.R
\name{scrape}
\alias{scrape}
\title{Scrape the content of authorized page/API}
\usage{
scrape(
  bow,
  query = NULL,
  params = NULL,
  accept = "html",
  content = NULL,
  verbose = FALSE
)
}
\arguments{
\item{bow}{host introduction object of class \code{polite}, \code{session} created by \code{bow()} or \code{nod()}}

\item{query}{named list of parameters to be appended to URL in the format \code{list(param1=valA, param2=valB)}}

\item{params}{deprecated. Use \code{query} argument above.}

\item{accept}{character value of expected data type to be returned by host (e.g. \code{html}, \code{json}, \code{xml}, \code{csv}, \code{txt}, etc.)}

\item{content}{MIME type (aka internet media type) used to override the content type returned by the server.
See http://en.wikipedia.org/wiki/Internet_media_type for a list of common types. You can add the \code{charset} parameter to override the server's default encoding}

\item{verbose}{extra feedback from the function. Defaults to \code{FALSE}}
}
\value{
Object of class \code{httr::response} which can be further processed by functions in \code{rvest} package
}
\description{
Scrape the content of authorized page/API
}
\examples{
\donttest{
 library(rvest)
  bow("https://en.wikipedia.org/wiki/List_of_cognitive_biases") \%>\%
   scrape(content="text/html; charset=UTF-8") \%>\%
   html_nodes(".wikitable") \%>\%
   html_table()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{guess_basename}
\alias{guess_basename}
\title{Guess download file name from the URL}
\usage{
guess_basename(x)
}
\arguments{
\item{x}{url to guess basename from}
}
\value{
guessed file name
}
\description{
Guess download file name from the URL
}
\examples{
guess_basename("https://bit.ly/polite_sticker")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/html.R
\name{html_attrs_dfr}
\alias{html_attrs_dfr}
\title{Convert collection of html nodes into data frame}
\usage{
html_attrs_dfr(
  x,
  attrs = NULL,
  trim = FALSE,
  defaults = NA_character_,
  add_text = TRUE
)
}
\arguments{
\item{x}{\code{xml_nodeset} object, containing text and attributes of interest}

\item{attrs}{character vector of attribute names. If missing, all attributes will be used}

\item{trim}{if \code{TRUE}, will trim leading and trailing spaces}

\item{defaults}{character vector of default values to be passed to \code{rvest::html_attr()}. Recycled to match length of \code{attrs}}

\item{add_text}{if \code{TRUE}, node content will be added as \code{.text} column (using \code{rvest::html_text})}
}
\value{
data frame with one row per xml node, consisting of an html_text column with text and additional columns with attributes
}
\description{
Convert collection of html nodes into data frame
}
\examples{
\donttest{
library(polite)
library(rvest)
bow("https://en.wikipedia.org/wiki/List_of_cognitive_biases") \%>\%
  scrape() \%>\%
  html_nodes("tr td:nth-child(1) a") \%>\%
  html_attrs_dfr()
  }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{null-coalesce}
\alias{null-coalesce}
\alias{\%otherwise\%}
\title{Null coalescing operator}
\usage{
lhs \%otherwise\% rhs
}
\description{
Null coalescing operator
}
\keyword{internal}
